@@ -68,7 +68,7 @@ public:
   /// Factor numbers <= y
   FactorTable(int64_t y, int threads)
   {
-    if (y > max())
+    if_unlikely(y > max())
       throw primecount_error("y must be <= FactorTable::max()");
 
     y = std::max<int64_t>(1, y);
@@ -84,7 +84,7 @@ public:
 
     int64_t sqrty = isqrt(y);
     int64_t thread_threshold = (int64_t) 1e7;
-    threads = ideal_num_threads(threads, y, thread_threshold);
+    threads = ideal_num_threads(y, threads, thread_threshold);
     int64_t thread_distance = ceil_div(y, threads);
     thread_distance += coprime_indexes_.size() - thread_distance % coprime_indexes_.size();
 
@@ -106,7 +106,7 @@ public:
         int64_t size = (to_index(high) + 1) - low_idx;
         std::fill_n(&factor_[low_idx], size, T_MAX);
 
-        int64_t start = first_coprime() - 1;
+        int64_t start = first_coprime();
         int64_t stop = high / first_coprime();
         primesieve::iterator it(start, stop);
 
