@@ -22,6 +22,13 @@
   #define __has_cpp_attribute(x) 0
 #endif
 
+#ifndef __has_include
+  #define __has_include(x) 0
+#endif
+
+// Required for std::unreachable()
+#include <utility>
+
 /// Enable expensive debugging assertions.
 /// These assertions enable e.g. bounds checks for the
 /// pod_vector and pod_array types.
@@ -89,6 +96,13 @@
   #define UNREACHABLE __builtin_unreachable()
 #elif defined(_MSC_VER)
   #define UNREACHABLE __assume(0)
+#elif __cplusplus >= 202301L && \
+      defined(__cpp_lib_unreachable)
+  // We prefer __builtin_unreachable() over std::unreachable()
+  // because GCC's std::unreachable() implementation uses
+  // __builtin_trap() instead of __builtin_unreachable() if
+  // _GLIBCXX_ASSERTIONS is defined.
+  #define UNREACHABLE std::unreachable()
 #else
   #define UNREACHABLE
 #endif
