@@ -29,6 +29,7 @@
 ///
 
 #include <LoadBalancerS2.hpp>
+#include <primecount-config.hpp>
 #include <primecount-internal.hpp>
 #include <StatusS2.hpp>
 #include <Sieve.hpp>
@@ -51,8 +52,7 @@ LoadBalancerS2::LoadBalancerS2(maxint_t x,
   is_print_(is_print),
   status_(x)
 {
-  if (threads > 1)
-    lock_.init();
+  lock_.init(threads);
 
   // The best performance is usually achieved using
   // a sieve array size that matches your CPU's L1
@@ -60,7 +60,7 @@ LoadBalancerS2::LoadBalancerS2(maxint_t x,
   // larger than your L1 cache size but smaller than
   // your L2 cache size (per core). Also, the
   // segment_size must be >= sqrt(sieve_limit).
-  int64_t sieve_bytes = 128 << 10;
+  int64_t sieve_bytes = L1D_CACHE_SIZE * 2;
   int64_t numbers_per_byte = 30;
   int64_t sqrt_limit = isqrt(sieve_limit);
   max_size_ = max(sieve_bytes * numbers_per_byte, sqrt_limit);
@@ -95,7 +95,7 @@ maxint_t LoadBalancerS2::get_sum() const
   return sum_;
 }
 
-bool LoadBalancerS2::get_work(ThreadSettings& thread)
+bool LoadBalancerS2::get_work(ThreadData& thread)
 {
   LockGuard lockGuard(lock_);
   sum_ += thread.sum;
@@ -122,7 +122,7 @@ bool LoadBalancerS2::get_work(ThreadSettings& thread)
   return is_work;
 }
 
-void LoadBalancerS2::update_load_balancing(const ThreadSettings& thread)
+void LoadBalancerS2::update_load_balancing(const ThreadData& thread)
 {
   if (thread.low > max_low_)
   {
@@ -159,7 +159,7 @@ void LoadBalancerS2::update_segment_size()
 /// Increase or decrease the number of segments per thread
 /// based on the remaining runtime.
 ///
-void LoadBalancerS2::update_number_of_segments(const ThreadSettings& thread)
+void LoadBalancerS2::update_number_of_segments(const ThreadData& thread)
 {
   // Near the end it is important that threads run only for
   // a short amount of time in order to ensure that all
@@ -182,10 +182,10 @@ void LoadBalancerS2::update_number_of_segments(const ThreadSettings& thread)
   // However for very large computations we want to further
   // reduce the thread runtimes in order to increase the
   // backup frequency. If the thread runtime is > 6 hours
-  // we reduce the thread runtime to about 50x the thread
+  // we reduce the thread runtime to about 200x the thread
   // initialization time.
   double init_secs = max(min_secs, thread.init_secs);
-  double init_factor = in_between(50, (3600 * 6) / init_secs, 5000);
+  double init_factor = in_between(200, (3600 * 6) / init_secs, 5000);
 
   // Reduce the thread runtime if it is much larger than
   // its initialization time. This increases the number of
