@@ -320,8 +320,15 @@ T AC_OpenMP(T x,
   T sum = 0;
   int64_t x13 = iroot<3>(x);
   int64_t sqrtx = isqrt(x);
+  int64_t xy = x / y;
+  int64_t xz = x / z;
+
+  // These load balancing settings work well on my
+  // dual-socket AMD EPYC 7642 server with 192 CPU cores.
   int64_t thread_threshold = 1000;
-  threads = ideal_num_threads(threads, x13, thread_threshold);
+  int max_threads = (int) std::pow(xz, 1 / 3.7);
+  threads = std::min(threads, max_threads);
+  threads = ideal_num_threads(x13, threads, thread_threshold);
   LoadBalancerAC loadBalancer(sqrtx, y, threads, is_print);
 
   // Initialize libdivide vector from primes vector
@@ -338,8 +345,8 @@ T AC_OpenMP(T x,
 
   int64_t pi_y = pi[y];
   int64_t pi_sqrtz = pi[isqrt(z)];
-  int64_t pi_root3_xy = pi[iroot<3>(x / y)];
-  int64_t pi_root3_xz = pi[iroot<3>(x / z)];
+  int64_t pi_root3_xy = pi[iroot<3>(xy)];
+  int64_t pi_root3_xz = pi[iroot<3>(xz)];
   RelaxedAtomic<int64_t> min_c1(max(k, pi_root3_xz) + 1);
 
   // In order to reduce the thread creation & destruction
@@ -360,6 +367,9 @@ T AC_OpenMP(T x,
     int64_t low, high;
 
     // C1 formula: pi[(x/z)^(1/3)] < b <= pi[pi_sqrtz]
+    // There are very few iterations in this loop,
+    // hence the use of an atomic loop counter (min_c1)
+    // won't cause any scaling issues.
     for (int64_t b = min_c1++; b <= pi_sqrtz; b = min_c1++)
     {
       int64_t prime = primes[b];
@@ -436,14 +446,16 @@ int64_t AC(int64_t x,
            int threads,
            bool is_print)
 {
+  double time;
+
   if (is_print)
   {
     print("");
     print("=== AC(x, y) ===");
     print_gourdon_vars(x, y, z, k, threads);
+    time = get_time();
   }
 
-  double time = get_time();
   int64_t x_star = get_x_star_gourdon(x, y);
   int64_t max_c_prime = y;
   int64_t max_a_prime = (int64_t) isqrt(x / x_star);
@@ -467,14 +479,16 @@ int128_t AC(int128_t x,
             int threads,
             bool is_print)
 {
+  double time;
+
   if (is_print)
   {
     print("");
     print("=== AC(x, y) ===");
     print_gourdon_vars(x, y, z, k, threads);
+    time = get_time();
   }
 
-  double time = get_time();
   int64_t x_star = get_x_star_gourdon(x, y);
   int64_t max_c_prime = y;
   int64_t max_a_prime = (int64_t) isqrt(x / x_star);
