@@ -74,7 +74,10 @@ X Phi0_OpenMP(X x,
               int64_t k,
               int threads)
 {
-  threads = ideal_num_threads(threads, y);
+  // These load balancing settings work well on my
+  // dual-socket AMD EPYC 7642 server with 192 CPU cores.
+  int64_t thread_threshold = (int64_t) 1e6;
+  threads = ideal_num_threads(y, threads, thread_threshold);
 
   auto primes = generate_primes<Y>(y);
   int64_t pi_y = primes.size() - 1;
@@ -101,14 +104,16 @@ int64_t Phi0(int64_t x,
              int threads,
              bool is_print)
 {
+  double time;
+
   if (is_print)
   {
     print("");
     print("=== Phi0(x, y) ===");
     print_gourdon_vars(x, y, z, k, threads);
+    time = get_time();
   }
 
-  double time = get_time();
   int64_t phi0 = Phi0_OpenMP(x, y, z, k, threads);
 
   if (is_print)
@@ -126,14 +131,16 @@ int128_t Phi0(int128_t x,
               int threads,
               bool is_print)
 {
+  double time;
+
   if (is_print)
   {
     print("");
     print("=== Phi0(x, y) ===");
     print_gourdon_vars(x, y, z, k, threads);
+    time = get_time();
   }
 
-  double time = get_time();
   int128_t phi0;
 
   // uses less memory
