@@ -54,7 +54,7 @@ int64_t S2_thread(int64_t x,
                   const pod_vector<int32_t>& primes,
                   const pod_vector<int32_t>& lpf,
                   const pod_vector<int32_t>& mu,
-                  ThreadSettings& thread)
+                  ThreadData& thread)
 {
   int64_t sum = 0;
   int64_t low = thread.low;
@@ -70,8 +70,8 @@ int64_t S2_thread(int64_t x,
   if (min_b > max_b)
     return 0;
 
-  Sieve sieve(low, segment_size, max_b);
   auto phi = generate_phi(low, max_b, primes, pi);
+  Sieve sieve(low, segment_size, max_b);
   thread.init_finished();
 
   // segmented sieve of Eratosthenes
@@ -174,21 +174,27 @@ int64_t S2(int64_t x,
            int threads,
            bool is_print)
 {
+  double time;
+
   if (is_print)
   {
     print("");
     print("=== S2(x, y) ===");
+    time = get_time();
   }
 
-  double time = get_time();
+  // These load balancing settings work well on my
+  // dual-socket AMD EPYC 7642 server with 192 CPU cores.
   int64_t thread_threshold = 1 << 20;
-  threads = ideal_num_threads(threads, z, thread_threshold);
+  int max_threads = (int) std::pow(z, 1 / 3.7);
+  threads = std::min(threads, max_threads);
+  threads = ideal_num_threads(z, threads, thread_threshold);
   LoadBalancerS2 loadBalancer(x, z, s2_approx, threads, is_print);
   PiTable pi(y, threads);
 
   #pragma omp parallel num_threads(threads)
   {
-    ThreadSettings thread;
+    ThreadData thread;
 
     while (loadBalancer.get_work(thread))
     {
@@ -236,12 +242,12 @@ int64_t pi_lmo_parallel(int64_t x,
     print(x, y, z, c, threads);
   }
 
-  int64_t p2 = P2(x, y, threads, is_print);
   auto primes = generate_primes<int32_t>(y);
   auto lpf = generate_lpf(y);
   auto mu = generate_moebius(y);
 
   int64_t pi_y = primes.size() - 1;
+  int64_t p2 = P2(x, y, pi_y, threads, is_print);
   int64_t s1 = S1(x, y, c, threads, is_print);
   int64_t s2_approx = S2_approx(x, pi_y, p2, s1);
   int64_t s2 = S2(x, y, z, c, s2_approx, primes, lpf, mu, threads, is_print);
