@@ -39,9 +39,7 @@
 #include <algorithm>
 #include <cmath>
 #include <utility>
-#include <vector>
 
-using std::vector;
 using namespace primecount;
 
 namespace {
@@ -94,10 +92,8 @@ public:
 
     // Make sure that there are no uninitialized
     // bits in the last sieve array element.
-    ASSERT(max_x_size_ > 0);
     max_x_ = max_x_size_ * 240 - 1;
     max_a_ = max_a;
-    sieve_.resize(max_a_ + 1);
   }
 
   /// Calculate phi(x, a) using the recursive formula:
@@ -112,27 +108,32 @@ public:
       return phi_tiny(x, a) * SIGN;
     else if (is_pix(x, a))
       return (pi_[x] - a + 1) * SIGN;
-    else if (is_cached(x, a))
+
+    // Cache small phi(x, i) results with i <= min(a, max_a_)
+    if (max_a_cached_ < min(a, max_a_) &&
+        (uint64_t) x <= max_x_)
+      init_cache(min(a, max_a_));
+
+    if (is_cached(x, a))
       return phi_cache(x, a) * SIGN;
 
-    // Cache all small phi(x, i) results with:
-    // x <= max_x && i <= min(a, max_a)
-    init_cache(x, a);
+    int64_t sum;
+    int64_t c = PhiTiny::max_a();
+    int64_t larger_c = min(max_a_cached_, a);
+    larger_c = std::max(c, larger_c);
+    ASSERT(c < a);
+
+    // Usually our algorithm starts at c because phi(x, c) can be
+    // computed in O(1) time using phi_tiny(x, c). However, if a
+    // larger value of c is cached, then it is better to start at that
+    // value, since phi_cache(x, larger_c) also takes O(1) time.
+    if (is_cached(x, larger_c))
+      sum = phi_cache(x, (c = larger_c)) * SIGN;
+    else
+      sum = phi_tiny(x, c) * SIGN;
 
     int64_t sqrtx = isqrt(x);
-    int64_t c = PhiTiny::get_c(sqrtx);
-    int64_t c_cached = min(a, max_a_cached_);
-    int64_t sum, i;
-
-    if (c >= c_cached ||
-        !is_cached(x, c_cached))
-      sum = phi_tiny(x, c) * SIGN;
-    else
-    {
-      c = c_cached;
-      ASSERT(c_cached <= a);
-      sum = phi_cache(x, c) * SIGN;
-    }
+    int64_t i;
 
     for (i = c + 1; i <= a; i++)
     {
@@ -142,26 +143,42 @@ public:
       // phi(x / prime[i], i - 1) = 1 even if x / prime[i] > prime[i-1].
       // This works because in this case there is no other prime
       // inside the interval ]prime[i-1], x / prime[i]].
-      if (primes_[i] > sqrtx)
-        break;
+      if_unlikely(primes_[i] > sqrtx)
+        goto phi_1;
+
       int64_t xp = fast_div(x, primes_[i]);
+
+      // All remaining loop iterations can be computed
+      // in O(1) time using the pi(x) lookup table.
       if (is_pix(xp, i - 1))
-        break;
-      sum += phi<-SIGN>(xp, i - 1);
+      {
+        sum += (pi_[xp] - i + 2) * -SIGN;
+        i += 1; break;
+      }
+
+      if (is_cached(xp, i - 1))
+        sum += phi_cache(xp, i - 1) * -SIGN;
+      else
+        sum += phi<-SIGN>(xp, i - 1);
     }
 
     for (; i <= a; i++)
     {
-      if (primes_[i] > sqrtx)
-        break;
-      int64_t xp = fast_div(x, primes_[i]);
-      // if a >= pi(sqrt(x)): phi(x, a) = pi(x) - a + 1
+      if_unlikely(primes_[i] > sqrtx)
+        goto phi_1;
+
+      // If a >= pi(sqrt(x)): phi(x, a) = pi(x) - a + 1
       // phi(xp, i - 1) = pi(xp) - (i - 1) + 1
       // phi(xp, i - 1) = pi(xp) - i + 2
+      int64_t xp = fast_div(x, primes_[i]);
+      ASSERT(is_pix(xp, i - 1));
       sum += (pi_[xp] - i + 2) * -SIGN;
     }
 
-    // phi(xp, i - 1) = 1 for i in [i, a]
+    phi_1:
+
+    // For i in ]pi(sqrt(x)), a]:
+    // phi(x / prime[i], i - 1) = 1
     sum += (a + 1 - i) * -SIGN;
     return sum;
   }
@@ -181,18 +198,20 @@ private:
   bool is_cached(uint64_t x, uint64_t a) const
   {
     return x <= max_x_ &&
-           a <= max_a_cached_;
+           a <= max_a_cached_ &&
+           a > PhiTiny::max_a();
   }
 
   int64_t phi_cache(uint64_t x, uint64_t a) const
   {
+    ASSERT(is_cached(x, a));
     uint64_t count = sieve_[a][x / 240].count;
     uint64_t bits = sieve_[a][x / 240].bits;
     uint64_t bitmask = unset_larger_[x % 240];
     return count + popcnt64(bits & bitmask);
   }
 
-  /// Cache phi(x, i) results with: x <= max_x && i <= min(a, max_a).
+  /// Cache phi(x, i) results with: x <= max_x && i <= a.
   /// Eratosthenes-like sieving algorithm that removes the first a primes
   /// and their multiples from the sieve array. Additionally this
   /// algorithm counts the numbers that are not divisible by any of the
@@ -200,57 +219,55 @@ private:
   /// counting has finished phi(x, a) results can be retrieved from the
   /// cache in O(1) using the phi_cache(x, a) method.
   ///
-  void init_cache(uint64_t x, uint64_t a)
+  void init_cache(uint64_t a)
   {
-    a = min(a, max_a_);
+    ASSERT(a > PhiTiny::max_a());
+    ASSERT(a <= max_a_);
 
-    if (x > max_x_ ||
-        a <= max_a_cached_)
-      return;
+    if (sieve_.empty())
+    {
+      ASSERT(max_a_ >= 3);
+      sieve_.resize(max_a_ + 1);
+      sieve_[3].resize(max_x_size_);
+      std::fill(sieve_[3].begin(), sieve_[3].end(), sieve_t{0, ~0ull});
+      max_a_cached_ = 3;
+    }
 
     uint64_t i = max_a_cached_ + 1;
+    ASSERT(a > max_a_cached_);
     max_a_cached_ = a;
-    i = max(i, 3);
 
     for (; i <= a; i++)
     {
+      // Initalize phi(x, i) with phi(x, i - 1)
+      if (i - 1 <= PhiTiny::max_a())
+        sieve_[i] = std::move(sieve_[i - 1]);
+      else
+      {
+        sieve_[i].resize(sieve_[i - 1].size());
+        std::copy(sieve_[i - 1].begin(), sieve_[i - 1].end(), sieve_[i].begin());
+      }
+
+      // Remove prime[i] and its multiples.
       // Each bit in the sieve array corresponds to an integer that
       // is not divisible by 2, 3 and 5. The 8 bits of each byte
       // correspond to the offsets { 1, 7, 11, 13, 17, 19, 23, 29 }.
-      if (i == 3)
-      {
-        sieve_[i].resize(max_x_size_);
-        std::fill(sieve_[i].begin(), sieve_[i].end(), sieve_t{0, ~0ull});
-      }
-      else
-      {
-        // Initalize phi(x, i) with phi(x, i - 1)
-        if (i - 1 <= PhiTiny::max_a())
-          sieve_[i] = std::move(sieve_[i - 1]);
-        else
-        {
-          sieve_[i].resize(sieve_[i - 1].size());
-          std::copy(sieve_[i - 1].begin(), sieve_[i - 1].end(), sieve_[i].begin());
-        }
+      uint64_t prime = primes_[i];
+      if (prime <= max_x_)
+        sieve_[i][prime / 240].bits &= unset_bit_[prime % 240];
+      for (uint64_t n = prime * prime; n <= max_x_; n += prime * 2)
+        sieve_[i][n / 240].bits &= unset_bit_[n % 240];
 
-        // Remove prime[i] and its multiples
-        uint64_t prime = primes_[i];
-        if (prime <= max_x_)
-          sieve_[i][prime / 240].bits &= unset_bit_[prime % 240];
-        for (uint64_t n = prime * prime; n <= max_x_; n += prime * 2)
-          sieve_[i][n / 240].bits &= unset_bit_[n % 240];
-
-        if (i > PhiTiny::max_a())
+      if (i > PhiTiny::max_a())
+      {
+        // Fill an array with the cumulative 1 bit counts.
+        // sieve[i][j] contains the count of numbers < j * 240 that
+        // are not divisible by any of the first i primes.
+        uint64_t count = 0;
+        for (auto& sieve : sieve_[i])
         {
-          // Fill an array with the cumulative 1 bit counts.
-          // sieve[i][j] contains the count of numbers < j * 240 that
-          // are not divisible by any of the first i primes.
-          uint64_t count = 0;
-          for (auto& sieve : sieve_[i])
-          {
-            sieve.count = (uint32_t) count;
-            count += popcnt64(sieve.bits);
-          }
+          sieve.count = (uint32_t) count;
+          count += popcnt64(sieve.bits);
         }
       }
     }
@@ -275,7 +292,7 @@ private:
   /// by any of the the first a primes. sieve[a][i].count
   /// contains the count of numbers < i * 240 that are not
   /// divisible by any of the first a primes.
-  vector<pod_vector<sieve_t>> sieve_;
+  pod_vector<pod_vector<sieve_t>> sieve_;
   const pod_vector<int32_t>& primes_;
   const PiTable& pi_;
 };
@@ -321,7 +338,7 @@ int64_t pix_upper(int64_t x)
 /// phi(x, a) counts the numbers <= x that are not divisible
 /// by any of the first a primes.
 ///
-int64_t phi_noprint(int64_t x, int64_t a, int threads)
+int64_t phi_OpenMP(int64_t x, int64_t a, int threads)
 {
   if (x < 1) return 0;
   if (a < 1) return x;
@@ -343,6 +360,9 @@ int64_t phi_noprint(int64_t x, int64_t a, int threads)
   if (a > pix_upper(sqrtx))
     return phi_pix(x, a, threads);
 
+  // We use a large pi(x) lookup table of size sqrt(x) to speed up our
+  // phi(x, a) implementation. As a drawback this increases the memory
+  // usage of our phi(x, a) implementation from O(a) to O(sqrt(x)).
   PiTable pi(sqrtx, threads);
   int64_t pi_sqrtx = pi[sqrtx];
 
@@ -355,10 +375,15 @@ int64_t phi_noprint(int64_t x, int64_t a, int threads)
     return phi_pix(x, a, threads);
 
   auto primes = generate_n_primes<int32_t>(a);
-  int64_t c = PhiTiny::get_c(sqrtx);
+  int64_t c = min(PhiTiny::max_a(), a);
   int64_t sum = phi_tiny(x, c);
+
+  // These load balancing settings work well on my
+  // dual-socket AMD EPYC 7642 server with 192 CPU cores.
   int64_t thread_threshold = (int64_t) 1e10;
-  threads = ideal_num_threads(threads, x, thread_threshold);
+  int max_threads = (int) std::sqrt(a);
+  threads = std::min(threads, max_threads);
+  threads = ideal_num_threads(x, threads, thread_threshold);
 
   #pragma omp parallel num_threads(threads) reduction(+: sum)
   {
@@ -387,14 +412,16 @@ int64_t phi(int64_t x,
             int threads,
             bool is_print)
 {
+  double time;
+
   if (is_print)
   {
     print("");
     print("=== phi(x, a) ===");
+    time = get_time();
   }
 
-  double time = get_time();
-  int64_t sum = phi_noprint(x, a, threads);
+  int64_t sum = phi_OpenMP(x, a, threads);
 
   if (is_print)
     print("phi", sum, time);
